# Copyright 2016 Quora, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from .futures import FutureBase
from . import async_task
from . import _debug


_debug_options = _debug.options

def result(value):
    """An async equivalent for return for async methods: result(x); return."""
    assert not isinstance(value, FutureBase), \
        "You probably forgot to yield this value before returning"
    raise async_task.AsyncTaskResult(value)

